#! /usr/bin/env python
"""Deduplication of remaining records in curated metadata repositories"""
from __future__ import annotations

import typing
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.dedupe

# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.DedupePackageEndpointInterface)
@dataclass
class CurationMissingDedupe(JsonSchemaMixin):
    """Deduplication of remaining records in a curated metadata repository"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        settings: dict,
    ):
        self.settings = self.settings_class.load_settings(data=settings)
        self.review_manager = dedupe_operation.review_manager

        self.__post_md_prepared_states = colrev.record.RecordState.get_post_x_states(
            state=colrev.record.RecordState.md_processed
        )

    def __create_dedupe_source_stats(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe
    ) -> None:
        # Note : reload to generate correct statistics

        Path("dedupe").mkdir(exist_ok=True)

        source_origins = [
            str(source.filename).replace("data/search/", "")
            for source in dedupe_operation.review_manager.settings.sources
        ]

        records = dedupe_operation.review_manager.dataset.load_records_dict()
        for source_origin in source_origins:
            selected_records = [
                r
                for r in records.values()
                if any(source_origin in co for co in r["colrev_origin"])
                and r["colrev_status"]
                in [
                    colrev.record.RecordState.md_prepared,
                    colrev.record.RecordState.md_needs_manual_preparation,
                    colrev.record.RecordState.md_imported,
                ]
            ]
            records_df = pd.DataFrame.from_records(list(selected_records))
            if records_df.shape[0] == 0:
                dedupe_operation.review_manager.logger.info(
                    f"{colors.GREEN}Source {source_origin} fully merged{colors.END}"
                )
            else:
                dedupe_operation.review_manager.logger.info(
                    f"{colors.ORANGE}Source {source_origin} not fully merged{colors.END}"
                )
                dedupe_operation.review_manager.logger.info(
                    f"Exporting details to dedupe/{source_origin}.xlsx"
                )

                records_df = records_df[
                    records_df.columns.intersection(
                        [
                            "ID",
                            "colrev_status",
                            "journal",
                            "booktitle",
                            "year",
                            "volume",
                            "number",
                            "title",
                            "author",
                        ]
                    )
                ]
                keys = list(
                    records_df.columns.intersection(["year", "volume", "number"])
                )
                if "year" in keys:
                    records_df.year = pd.to_numeric(records_df.year, errors="coerce")
                if "volume" in keys:
                    records_df.volume = pd.to_numeric(
                        records_df.volume, errors="coerce"
                    )
                if "number" in keys:
                    records_df.number = pd.to_numeric(
                        records_df.number, errors="coerce"
                    )
                records_df.sort_values(by=keys, inplace=True)
                records_df.to_excel(f"dedupe/{source_origin}.xlsx", index=False)

    def __get_same_toc_recs(
        self, *, record: colrev.record.Record, records: dict
    ) -> list:
        if self.review_manager.force_mode:
            if record.data["colrev_status"] in self.__post_md_prepared_states:
                return []
        else:
            # only dedupe md_prepared records
            if record.data["colrev_status"] not in [
                colrev.record.RecordState.md_prepared
            ]:
                return []
        if record.data.get("title", "") == "":
            return []

        try:
            toc_key = record.get_toc_key()
        except colrev_exceptions.NotTOCIdentifiableException:
            return []

        same_toc_recs = []
        for record_candidate in records.values():
            try:
                candidate_toc_key = colrev.record.Record(
                    data=record_candidate
                ).get_toc_key()
            except colrev_exceptions.NotTOCIdentifiableException:
                continue
            if toc_key != candidate_toc_key:
                continue
            if record_candidate["ID"] == record.data["ID"]:
                continue
            if record_candidate["colrev_status"] in [
                colrev.record.RecordState.md_prepared,
                colrev.record.RecordState.md_needs_manual_preparation,
                colrev.record.RecordState.md_imported,
            ]:
                continue
            same_toc_recs.append(record_candidate)
        return same_toc_recs

    def __print_same_toc_recs(
        self, *, same_toc_recs: list, record: colrev.record.Record
    ) -> list:
        for same_toc_rec in same_toc_recs:
            same_toc_rec["similarity"] = colrev.record.PrepRecord.get_record_similarity(
                record_a=colrev.record.Record(data=same_toc_rec), record_b=record
            )

        same_toc_recs = sorted(
            same_toc_recs, key=lambda d: d["similarity"], reverse=True
        )
        if len(same_toc_recs) > 20:
            same_toc_recs = same_toc_recs[0:20]

        i = 0
        for i, same_toc_rec in enumerate(same_toc_recs):
            author_title_string = (
                f"{same_toc_rec.get('author', 'NO_AUTHOR')} : "
                + f"{same_toc_rec.get('title', 'NO_TITLE')}"
            )

            if same_toc_rec["similarity"] > 0.8:
                print(f"{i + 1} - {colors.ORANGE}{author_title_string}{colors.END}")

            else:
                print(f"{i + 1} - {author_title_string}")
        return same_toc_recs

    def __get_nr_recs_to_merge(self, *, records: dict) -> int:
        if self.review_manager.force_mode:
            self.review_manager.logger.info(
                "Scope: md_prepared, md_needs_manual_preparation, md_imported"
            )
            nr_recs_to_merge = len(
                [
                    x
                    for x in records.values()
                    if x["colrev_status"] not in self.__post_md_prepared_states
                ]
            )
        else:
            self.review_manager.logger.info("Scope: md_prepared")
            nr_recs_to_merge = len(
                [
                    x
                    for x in records.values()
                    if x["colrev_status"] in [colrev.record.RecordState.md_prepared]
                ]
            )
        return nr_recs_to_merge

    def __process_missing_duplicates(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe
    ) -> dict:
        records = dedupe_operation.review_manager.dataset.load_records_dict()

        nr_recs_to_merge = self.__get_nr_recs_to_merge(records=records)

        nr_recs_checked = 0
        results: typing.Dict[str, list] = {
            "decision_list": [],
            "add_records_to_md_processed_list": [],
            "records_to_prepare": [],
        }

        for record_dict in records.values():
            record = colrev.record.Record(data=record_dict)

            same_toc_recs = self.__get_same_toc_recs(record=record, records=records)

            if len(same_toc_recs) == 0:
                print("no same toc records")
                continue

            print("\n\n\n")
            print(colors.ORANGE)
            record.print_citation_format()
            print(colors.END)

            same_toc_recs = self.__print_same_toc_recs(
                same_toc_recs=same_toc_recs, record=record
            )
            i = len(same_toc_recs)
            valid_selection = False
            quit_pressed = False
            while not valid_selection:
                ret = input(
                    f"({nr_recs_checked}/{nr_recs_to_merge}) "
                    f"Merge with record [{1}...{i+1} / s / a / p / q]?   "
                )
                if ret == "s":
                    valid_selection = True
                elif ret == "q":
                    quit_pressed = True
                    valid_selection = True
                elif ret == "a":
                    results["add_records_to_md_processed_list"].append(
                        record.data["ID"]
                    )
                    valid_selection = True
                elif ret == "p":
                    results["records_to_prepare"].append(record.data["ID"])
                    valid_selection = True
                elif ret.isdigit():
                    if int(ret) - 1 <= i:
                        rec2 = same_toc_recs[int(ret) - 1]
                        if record.data["colrev_status"] < rec2["colrev_status"]:
                            results["decision_list"].append(
                                {
                                    "ID1": rec2["ID"],
                                    "ID2": record.data["ID"],
                                    "decision": "duplicate",
                                }
                            )
                        else:
                            results["decision_list"].append(
                                {
                                    "ID1": record.data["ID"],
                                    "ID2": rec2["ID"],
                                    "decision": "duplicate",
                                }
                            )

                        valid_selection = True
            nr_recs_checked += 1
            if quit_pressed:
                break
        return results

    def run_dedupe(self, dedupe_operation: colrev.ops.dedupe.Dedupe) -> None:
        """Run the dedupe procedure for remaining records in curations"""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        # export sets of non-merged records
        # (and merged records a different xlsx for easy sort/merge)

        # Note : this script is necessary because the active learning is insufficient:
        # the automated ML-deduplication still has a certain error rate
        # which makes it less useful for curations
        # the active learning labeling presents cases on both sides
        # (likely duplicates and non-duplicates to maximize training quality)
        # For the curation, we are only interested in the duplicate, not the classifier

        print("\n\n")
        print(
            "In the following, "
            "records can be added to the curated (md_processed*) records.\n"
            "Curated records are displayed for the same table-of-content item "
            "(i.e., same year/volume/number)"
        )
        print("\n\n")

        ret = self.__process_missing_duplicates(dedupe_operation=dedupe_operation)

        if len(ret["decision_list"]) > 0:
            print("Duplicates identified:")
            print(ret["decision_list"])
            preferred_masterdata_sources = [
                s
                for s in dedupe_operation.review_manager.settings.sources
                if s.endpoint != "colrev.pdfs_dir"
            ]

            dedupe_operation.apply_merges(
                results=ret["decision_list"],
                preferred_masterdata_sources=preferred_masterdata_sources,
            )

        if len(ret["records_to_prepare"]) > 0:
            records = dedupe_operation.review_manager.dataset.load_records_dict()
            for record_id, record_dict in records.items():
                if record_id in ret["records_to_prepare"]:
                    record = colrev.record.Record(data=record_dict)
                    record.set_status(
                        target_state=colrev.record.RecordState.md_needs_manual_preparation
                    )

            dedupe_operation.review_manager.dataset.save_records_dict(records=records)

        if len(ret["decision_list"]) > 0 or len(ret["records_to_prepare"]) > 0:
            dedupe_operation.review_manager.dataset.add_record_changes()

            dedupe_operation.review_manager.create_commit(
                msg="Merge duplicate records",
            )

        if len(ret["add_records_to_md_processed_list"]) > 0:
            records = dedupe_operation.review_manager.dataset.load_records_dict()
            for record_id, record_dict in records.items():
                if record_id in ret["add_records_to_md_processed_list"]:
                    if record_dict["colrev_status"] in [
                        colrev.record.RecordState.md_prepared,
                        colrev.record.RecordState.md_needs_manual_preparation,
                        colrev.record.RecordState.md_imported,
                    ]:
                        record = colrev.record.Record(data=record_dict)
                        record.set_status(
                            target_state=colrev.record.RecordState.md_processed
                        )

            dedupe_operation.review_manager.dataset.save_records_dict(records=records)
            dedupe_operation.review_manager.dataset.add_record_changes()

            input("Edit records (if any) and press Enter")

            dedupe_operation.review_manager.dataset.add_record_changes()

            dedupe_operation.review_manager.create_commit(
                msg="Add non-duplicate records",
            )

        self.__create_dedupe_source_stats(dedupe_operation=dedupe_operation)



#! /usr/bin/env python
"""Adding of journal rankings to metadata"""
from __future__ import annotations

from dataclasses import dataclass

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.local_index
import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.prep


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class AddJournalRanking(JsonSchemaMixin):
    """Class for add _journal_ranking"""

    settings_class = colrev.env.package_manager.DefaultSettings

    source_correction_hint = "check with the developer"
    always_apply_changes = False
    ci_supported: bool = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.local_index = colrev.env.local_index.LocalIndex()

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Add Journalranking to Metadata"""

        if record.data.get("journal", "") == "":
            return record

        rankings = self.local_index.search_in_database(record.data["journal"])
        # extend: include journal-impact factor or ranking category
        if rankings:
            rankings_str = ",".join(r["ranking"] for r in rankings)
        else:
            rankings_str = "not included in a ranking"

        record.update_field(
            key="journal_ranking",
            value=rankings_str,
            source="add_journal_ranking",
            note="",
        )

        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Checkers for CoLRev repositories"""
from __future__ import annotations

import os
import re
import sys
import typing
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

import yaml
from git.exc import InvalidGitRepositoryError

import colrev.exceptions as colrev_exceptions
import colrev.operation
from colrev.exit_codes import ExitCodes

if TYPE_CHECKING:
    import colrev.review_manager


class Checker:
    """The CoLRev checker makes sure the project setup is ok"""

    records: typing.Dict[str, typing.Any] = {}

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        self.review_manager = review_manager

        self.review_manager.notified_next_operation = (
            colrev.operation.OperationsType.check
        )

    def get_colrev_versions(self) -> list[str]:
        """Get the colrev version as a list: (last_version, current_version)"""
        current_colrev_version = version("colrev")
        last_colrev_version = current_colrev_version
        last_colrev_version = self.review_manager.settings.project.colrev_version
        if last_colrev_version.endswith("."):
            last_colrev_version += "0"
        return [last_colrev_version, current_colrev_version]

    def __check_software(self) -> None:
        last_version, current_version = self.get_colrev_versions()
        if last_version != current_version:
            raise colrev_exceptions.CoLRevUpgradeError(last_version, current_version)
        if not sys.version_info > (2, 7):
            raise colrev_exceptions.CoLRevException("CoLRev does not support Python 2.")
        if sys.version_info < (3, 5):
            self.review_manager.logger.warning(
                "CoLRev uses Python 3.8 features (currently, %s is installed). Please upgrade.",
                sys.version_info,
            )

    def check_repository_setup(self) -> None:
        """Check the repository setup"""

        # 1. git repository?
        if not self.__is_git_repo():
            raise colrev_exceptions.RepoSetupError()

        # 2. colrev project?
        if not self.__is_colrev_project():
            raise colrev_exceptions.RepoSetupError(
                "No colrev repository."
                + "To retrieve a shared repository, use colrev init."
                + "To initalize a new repository, "
                + "execute the command in an empty directory."
            )

        # 3. Pre-commit hooks installed?
        self.__require_colrev_hooks_installed()

    @classmethod
    def in_virtualenv(cls) -> bool:
        """Check whether CoLRev operates in a virtual environment"""

        def get_base_prefix_compat() -> str:
            return (
                getattr(sys, "base_prefix", None)
                or getattr(sys, "real_prefix", None)
                or sys.prefix
            )

        return get_base_prefix_compat() != sys.prefix

    def __check_git_conflicts(self) -> None:
        # Note: when check is called directly from the command line.
        # pre-commit hooks automatically notify on merge conflicts

        git_repo = self.review_manager.dataset.get_repo()
        unmerged_blobs = git_repo.index.unmerged_blobs()

        for path, list_of_blobs in unmerged_blobs.items():
            for stage, _ in list_of_blobs:
                if stage != 0:
                    raise colrev_exceptions.GitConflictError(Path(path))

    def __is_git_repo(self) -> bool:
        try:
            if not (self.review_manager.path / Path(".git")).is_dir():
                return False
            _ = self.review_manager.dataset.get_repo().git_dir
            return True
        except InvalidGitRepositoryError:
            return False

    def __is_colrev_project(self) -> bool:
        required_paths = [
            Path(".pre-commit-config.yaml"),
            Path(".gitignore"),
            Path("settings.json"),
        ]
        if not all((self.review_manager.path / x).is_file() for x in required_paths):
            return False
        return True

    def __get_installed_hooks(self) -> list:
        installed_hooks = []
        with open(".pre-commit-config.yaml", encoding="utf8") as pre_commit_y:
            pre_commit_config = yaml.load(pre_commit_y, Loader=yaml.SafeLoader)
        for repository in pre_commit_config["repos"]:
            installed_hooks.extend([hook["id"] for hook in repository["hooks"]])
        return installed_hooks

    def __require_colrev_hooks_installed(self) -> bool:
        required_hooks = [
            "colrev-hooks-check",
            "colrev-hooks-format",
            "colrev-hooks-report",
            "colrev-hooks-share",
        ]
        installed_hooks = self.__get_installed_hooks()
        hooks_activated = set(required_hooks).issubset(set(installed_hooks))
        if not hooks_activated:
            missing_hooks = [x for x in required_hooks if x not in installed_hooks]
            raise colrev_exceptions.RepoSetupError(
                f"missing hooks in .pre-commit-config.yaml ({', '.join(missing_hooks)})"
            )

        if not self.review_manager.in_ci_environment():
            pch_file = Path(".git/hooks/pre-commit")
            if pch_file.is_file():
                with open(pch_file, encoding="utf8") as file:
                    if "File generated by pre-commit" not in file.read(4096):
                        raise colrev_exceptions.RepoSetupError(
                            "pre-commit hooks not installed (use pre-commit install)"
                        )
            else:
                raise colrev_exceptions.RepoSetupError(
                    "pre-commit hooks not installed (use pre-commit install)"
                )

            psh_file = Path(".git/hooks/pre-push")
            if psh_file.is_file():
                with open(psh_file, encoding="utf8") as file:
                    if "File generated by pre-commit" not in file.read(4096):
                        raise colrev_exceptions.RepoSetupError(
                            "pre-commit push hooks not installed "
                            "(use pre-commit install --hook-type pre-push)"
                        )
            else:
                raise colrev_exceptions.RepoSetupError(
                    "pre-commit push hooks not installed "
                    "(use pre-commit install --hook-type pre-push)"
                )

            pcmh_file = Path(".git/hooks/prepare-commit-msg")
            if pcmh_file.is_file():
                with open(pcmh_file, encoding="utf8") as file:
                    if "File generated by pre-commit" not in file.read(4096):
                        raise colrev_exceptions.RepoSetupError(
                            "pre-commit prepare-commit-msg hooks not installed "
                            "(use pre-commit install --hook-type prepare-commit-msg)"
                        )
            else:
                raise colrev_exceptions.RepoSetupError(
                    "pre-commit prepare-commit-msg hooks not installed "
                    "(use pre-commit install --hook-type prepare-commit-msg)"
                )

        return True

    def __retrieve_ids_from_bib(self, *, file_path: Path) -> list:
        assert file_path.suffix == ".bib"
        record_ids = []
        with open(file_path, encoding="utf8") as file:
            line = file.readline()
            while line:
                if "@" in line[:5]:
                    record_id = line[line.find("{") + 1 : line.rfind(",")]
                    record_ids.append(record_id.lstrip())
                line = file.readline()
        return record_ids

    def __check_colrev_origins(self, *, status_data: dict) -> None:
        """Check colrev_origins"""

        # Check whether each record has an origin
        if not len(status_data["entries_without_origin"]) == 0:
            raise colrev_exceptions.OriginError(
                f"Entries without origin: {', '.join(status_data['entries_without_origin'])}"
            )

        if (
            self.review_manager.dataset.records_changed()
            or self.review_manager.verbose_mode
        ):
            # Check for broken origins
            all_record_links = []
            for bib_file in self.review_manager.search_dir.glob("*.bib"):
                self.review_manager.logger.debug(bib_file)
                search_ids = self.__retrieve_ids_from_bib(file_path=bib_file)
                for search_id in search_ids:
                    all_record_links.append(bib_file.name + "/" + search_id)
            delta = set(status_data["record_links_in_bib"]) - set(all_record_links)
            if len(delta) > 0:
                raise colrev_exceptions.OriginError(f"broken origins: {delta}")

        # Check for non-unique origins
        non_unique_origins = []
        for origin, record_ids in status_data["origin_ID_list"].items():
            if len(record_ids) > 1:
                if not origin.startswith("md_"):
                    non_unique_origins.append(f"{origin} - {','.join(record_ids)}")
        if non_unique_origins:
            raise colrev_exceptions.OriginError(
                f'Non-unique origins: {" , ".join(set(non_unique_origins))}'
            )

    def check_fields(self, *, status_data: dict) -> None:
        """Check field values"""

        # Check status fields
        status_schema = colrev.record.RecordState
        stat_diff = set(status_data["status_fields"]).difference(status_schema)
        if stat_diff:
            raise colrev_exceptions.FieldValueError(
                f"status field(s) {stat_diff} not in {status_schema}"
            )

    def check_status_transitions(self, *, status_data: dict) -> None:
        """Check for invalid state transitions"""
        # Note : currently, we do not prevent particular transitions.
        # We may decide to provide settings parameters to apply
        # more restrictive rules related to valid transitions.

        # We allow particular combinations of multiple transitions
        # if len(set(status_data["start_states"])) > 1:
        #     raise colrev_exceptions.StatusTransitionError(
        #         "multiple transitions from different "
        #         f'start states ({set(status_data["start_states"])})'
        #     )

        # We may apply more restrictive criteria to prevent invalid_state_transitions
        # E.g., setting a record from rev_synthesized to rev_included should be ok.
        # if len(set(status_data["invalid_state_transitions"])) > 0:
        #     raise colrev_exceptions.StatusTransitionError(
        #         "invalid state transitions: \n    "
        #         + "\n    ".join(status_data["invalid_state_transitions"])
        #     )

    def __check_individual_record_screen(
        self,
        *,
        record_id: str,
        status: colrev.record.RecordState,
        screen_crit: str,
        field_errors: typing.List[str],
        pattern: str,
        pattern_inclusion: str,
        criteria: typing.List[str],
    ) -> None:
        # No screening criteria allowed before screen
        if (
            status
            not in colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.rev_included
            )
            and status != colrev.record.RecordState.md_needs_manual_preparation
        ):
            if "NA" != screen_crit:
                raise colrev_exceptions.FieldValueError(
                    f"{record_id}: screen_crit != NA ({screen_crit})"
                )
            return

        # All screening criteria must match pattern
        if not re.match(pattern, screen_crit):
            # Note: this should also catch cases of missing
            # screening criteria
            field_errors.append(
                "Screening criteria field not matching "
                f"pattern: {screen_crit} ({record_id}; criteria: {criteria})"
            )
            return

        # Included papers must match inclusion pattern
        if status in [
            colrev.record.RecordState.rev_included,
            colrev.record.RecordState.rev_synthesized,
        ]:
            if not re.match(pattern_inclusion, screen_crit):
                field_errors.append(
                    "Included record with screening_criterion satisfied: "
                    f"{record_id}, {status}, {screen_crit}"
                )
            return

        # Excluded papers must match exclusion pattern
        if status == colrev.record.RecordState.rev_excluded:
            if ["NA"] == criteria:
                if screen_crit == "NA":
                    return
                field_errors.append(f"screen_crit field not NA: {screen_crit}")

            if "=out" not in screen_crit:
                self.review_manager.logger.error("criteria: %s", criteria)
                field_errors.append(
                    "Excluded record with no screening_criterion violated: "
                    f"{record_id}, {status}, {screen_crit}"
                )

    def __check_records_screen(self, *, status_data: dict) -> None:
        """Check consistency of screening criteria and status"""

        if not status_data["screening_criteria_list"]:
            return

        field_errors: typing.List[str] = []

        screening_criteria = self.review_manager.settings.screen.criteria
        if not screening_criteria:
            criteria = ["NA"]
            pattern = "^NA$"
            pattern_inclusion = "^NA$"
        else:
            pattern = "=(in|out);".join(screening_criteria.keys()) + "=(in|out)"
            pattern_inclusion = "=in;".join(screening_criteria.keys()) + "=in"
            criteria = list(screening_criteria.keys())

        for [record_id, status, screen_crit] in status_data["screening_criteria_list"]:
            self.__check_individual_record_screen(
                record_id=record_id,
                status=status,
                screen_crit=screen_crit,
                field_errors=field_errors,
                pattern=pattern,
                pattern_inclusion=pattern_inclusion,
                criteria=criteria,
            )

        if field_errors:
            raise colrev_exceptions.FieldValueError(
                "\n    " + "\n    ".join(field_errors)
            )

    def __check_change_in_propagated_id_in_file(
        self,
        *,
        notifications: list,
        root: str,
        filename: str,
        prior_id: str,
        new_id: str,
    ) -> None:
        if prior_id == str(Path(filename).name):
            msg = (
                f"Old ID ({prior_id}, changed to {new_id} in the "
                + f"RECORDS_FILE) found in filepath: {filename}"
            )
            if msg not in notifications:
                notifications.append(msg)

        # self.review_manager.logger.debug("Checking %s", name)
        if filename.endswith(".bib"):
            retrieved_ids = self.__retrieve_ids_from_bib(
                file_path=Path(os.path.join(root, filename))
            )
            if prior_id in retrieved_ids:
                msg = (
                    f"Old ID ({prior_id}, changed to {new_id} in "
                    + f"the RECORDS_FILE) found in file: {filename}"
                )
                if msg not in notifications:
                    notifications.append(msg)
        else:
            with open(os.path.join(root, filename), encoding="utf8") as file:
                line = file.readline()
                while line:
                    if filename.endswith(".bib") and "@" in line[:5]:
                        line = file.readline()
                    if prior_id in line:
                        msg = (
                            f"Old ID ({prior_id}, to {new_id} in "
                            + f"the RECORDS_FILE) found in file: {filename}"
                        )
                        if msg not in notifications:
                            notifications.append(msg)
                    line = file.readline()

    def check_change_in_propagated_id(
        self, *, prior_id: str, new_id: str = "TBD", project_context: Path
    ) -> list:
        """Check whether propagated IDs were changed

        A propagated ID is a record ID that is stored outside the records.bib.
        Propagated IDs should not be changed in the records.bib
        because this would break the link between the propagated ID and its metadata.
        """

        ignore_patterns = [
            ".git",
            ".report.log",
            ".pre-commit-config.yaml",
            "data/search",
            "records.bib",
        ]

        text_formats = [".txt", ".csv", ".md", ".bib", ".yaml"]
        notifications: typing.List[str] = []
        for root, dirs, files in os.walk(project_context, topdown=False):
            for filename in files:
                if any(
                    (x in filename) or (x in root) for x in ignore_patterns
                ) or not any(filename.endswith(x) for x in text_formats):
                    # self.review_manager.logger.debug("Skipping %s", name)
                    continue
                self.__check_change_in_propagated_id_in_file(
                    notifications=notifications,
                    root=root,
                    filename=filename,
                    prior_id=prior_id,
                    new_id=new_id,
                )

            for dir_name in dirs:
                if any((x in dir_name) or (x in root) for x in ignore_patterns):
                    continue
                if prior_id in dir_name:
                    notifications.append(
                        f"Old ID ({prior_id}, changed to {new_id} in the "
                        f"RECORDS_FILE) found in filepath: {dir_name}"
                    )
        return notifications

    def __check_change_in_propagated_ids(
        self, *, prior: dict, status_data: dict
    ) -> None:
        """Check for changes in propagated IDs"""

        if "persisted_IDs" not in prior:
            return
        for prior_origin, prior_id in prior["persisted_IDs"]:
            if prior_origin not in status_data["origin_ID_list"]:
                # Note: this does not catch origins removed before md_processed
                raise colrev_exceptions.OriginError(f"origin removed: {prior_origin}")
            new_ids = status_data["origin_ID_list"][prior_origin]
            if prior_id not in new_ids:
                notifications = self.check_change_in_propagated_id(
                    prior_id=prior_id,
                    new_id=",".join(new_ids),
                    project_context=self.review_manager.path,
                )
                notifications.append(
                    "ID of processed record changed from "
                    f"{prior_id} to {','.join(new_ids)}"
                )
                raise colrev_exceptions.PropagatedIDChange(notifications)

    def check_sources(self) -> None:
        """Check the sources"""
        for source in self.review_manager.settings.sources:
            if not source.filename.is_file():
                self.review_manager.logger.debug(
                    f"Search details without file: {source.filename}"
                )

    def __retrieve_prior(self) -> dict:
        prior: dict = {"colrev_status": [], "persisted_IDs": []}
        prior_records = next(self.review_manager.dataset.load_records_from_history())
        for prior_record in prior_records.values():
            for orig in prior_record["colrev_origin"]:
                prior["colrev_status"].append([orig, prior_record["colrev_status"]])
                if prior_record[
                    "colrev_status"
                ] in colrev.record.RecordState.get_post_x_states(
                    state=colrev.record.RecordState.md_processed
                ):
                    prior["persisted_IDs"].append([orig, prior_record["ID"]])
        return prior

    def __get_status_transitions(
        self,
        *,
        record_id: str,
        origin: list,
        prior: dict,
        status: colrev.record.RecordState,
        status_data: dict,
    ) -> dict:
        prior_status = []
        if "colrev_status" in prior:
            prior_status = [
                stat for (org, stat) in prior["colrev_status"] if org in origin
            ]

        status_transition = {}
        if len(prior_status) == 0:
            status_transition[record_id] = "load"
        else:
            proc_transition_list: list = [
                x["trigger"]
                for x in colrev.record.RecordStateModel.transitions
                if str(x["source"]) == prior_status[0] and str(x["dest"]) == status
            ]
            if len(proc_transition_list) == 0 and prior_status[0] != status:
                status_data["start_states"].append(prior_status[0])
                if prior_status[0] not in colrev.record.RecordState:
                    raise colrev_exceptions.StatusFieldValueError(
                        record_id, "colrev_status", prior_status[0]
                    )
                if status not in colrev.record.RecordState:
                    raise colrev_exceptions.StatusFieldValueError(
                        record_id, "colrev_status", str(status)
                    )

                status_data["invalid_state_transitions"].append(
                    f"{record_id}: {prior_status[0]} to {status}"
                )
            if 0 == len(proc_transition_list):
                status_transition[record_id] = "load"
            else:
                proc_transition = proc_transition_list.pop()
                status_transition[record_id] = proc_transition
        return status_transition

    def __retrieve_status_data(self, *, prior: dict, records: dict) -> dict:
        status_data: dict = {
            "pdf_not_exists": [],
            "status_fields": [],
            "status_transitions": [],
            "start_states": [],
            "screening_criteria_list": [],
            "IDs": [],
            "entries_without_origin": [],
            "record_links_in_bib": [],
            "persisted_IDs": [],
            "origin_ID_list": {},
            "invalid_state_transitions": [],
        }

        for record_dict in records.values():
            status_data["IDs"].append(record_dict["ID"])

            for org in record_dict["colrev_origin"]:
                if org in status_data["origin_ID_list"]:
                    status_data["origin_ID_list"][org].append(record_dict["ID"])
                else:
                    status_data["origin_ID_list"][org] = [record_dict["ID"]]

            post_md_processed_states = colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.md_processed
            )
            if record_dict["colrev_status"] in post_md_processed_states:
                for origin_part in record_dict["colrev_origin"]:
                    status_data["persisted_IDs"].append(
                        [origin_part, record_dict["ID"]]
                    )

            if "file" in record_dict:
                if Path(record_dict["file"]).is_file():
                    status_data["pdf_not_exists"].append(record_dict["ID"])

            if [] != record_dict.get("colrev_origin", []):
                for org in record_dict["colrev_origin"]:
                    status_data["record_links_in_bib"].append(org)
            else:
                status_data["entries_without_origin"].append(record_dict["ID"])

            status_data["status_fields"].append(record_dict["colrev_status"])

            if "screening_criteria" in record_dict:
                ec_case = [
                    record_dict["ID"],
                    record_dict["colrev_status"],
                    record_dict["screening_criteria"],
                ]
                status_data["screening_criteria_list"].append(ec_case)

            status_transition = self.__get_status_transitions(
                record_id=record_dict["ID"],
                origin=record_dict["colrev_origin"],
                prior=prior,
                status=record_dict["colrev_status"],
                status_data=status_data,
            )

            status_data["status_transitions"].append(status_transition)

        return status_data

    def check_repo_basics(self) -> list:
        """Calls data.main() to update the stats"""

        data_operation = self.review_manager.get_data_operation(
            notify_state_transition_operation=False
        )

        if self.review_manager.dataset.records_file.is_file():
            self.records = self.review_manager.dataset.load_records_dict()

        check_scripts: list[dict[str, typing.Any]] = []
        data_checks = [
            {
                "script": data_operation.main,
                "params": {"records": self.records, "silent_mode": True},
            },
            {
                "script": self.review_manager.update_status_yaml,
                "params": {"records": self.records},
            },
        ]

        check_scripts.extend(data_checks)

        failure_items = []
        for check_script in check_scripts:
            try:
                # self.review_manager.logger.info(check_script["script"])
                if not check_script["params"]:
                    # self.review_manager.logger.debug(
                    #     "%s() called", check_script["script"].__name__
                    # )
                    check_script["script"]()
                else:
                    # self.review_manager.logger.debug(
                    #     "%s(params) called", check_script["script"].__name__
                    # )
                    if isinstance(check_script["params"], list):
                        check_script["script"](*check_script["params"])
                    else:
                        check_script["script"](**check_script["params"])
                # self.review_manager.logger.debug(
                #     "%s: passed\n", check_script["script"].__name__
                # )
            except (
                colrev_exceptions.MissingDependencyError,
                colrev_exceptions.GitConflictError,
                colrev_exceptions.PropagatedIDChange,
                colrev_exceptions.DuplicateIDsError,
                colrev_exceptions.OriginError,
                colrev_exceptions.FieldValueError,
                colrev_exceptions.StatusTransitionError,
                colrev_exceptions.UnstagedGitChangesError,
                colrev_exceptions.StatusFieldValueError,
            ) as exc:
                failure_items.append(f"{type(exc).__name__}: {exc}")
        return failure_items

    def check_repo_extended(self) -> list:
        """Calls all checks that require prior data (take longer)"""

        # pylint: disable=not-a-mapping

        self.records: typing.Dict[str, typing.Any] = {}
        if self.review_manager.dataset.records_file.is_file():
            self.records = self.review_manager.dataset.load_records_dict()

        # We work with exceptions because each issue may be raised in different checks.
        # Currently, linting is limited for the scripts.

        environment_manager = self.review_manager.get_environment_manager()
        check_scripts: list[dict[str, typing.Any]] = [
            {
                "script": environment_manager.check_git_installed,
                "params": [],
            },
            {"script": self.__check_git_conflicts, "params": []},
            {"script": self.check_repository_setup, "params": []},
            {"script": self.__check_software, "params": []},
        ]

        if self.review_manager.dataset.records_file.is_file():
            if self.review_manager.dataset.file_in_history(
                filepath=self.review_manager.dataset.RECORDS_FILE_RELATIVE
            ):
                prior = self.__retrieve_prior()
                self.review_manager.logger.debug("prior")
                self.review_manager.logger.debug(
                    self.review_manager.p_printer.pformat(prior)
                )
            else:  # if RECORDS_FILE not yet in git history
                prior = {}

            status_data = self.__retrieve_status_data(prior=prior, records=self.records)

            main_refs_checks = [
                {"script": self.check_sources, "params": []},
            ]
            # Note : duplicate record IDs are already prevented by pybtex...

            if prior:  # if RECORDS_FILE in git history
                main_refs_checks.extend(
                    [
                        {
                            "script": self.__check_colrev_origins,
                            "params": {"status_data": status_data},
                        },
                        {
                            "script": self.__check_change_in_propagated_ids,
                            "params": {"prior": prior, "status_data": status_data},
                        },
                        {
                            "script": self.check_status_transitions,
                            "params": {"status_data": status_data},
                        },
                        {
                            "script": self.__check_records_screen,
                            "params": {"status_data": status_data},
                        },
                        {
                            "script": self.check_fields,
                            "params": {"status_data": status_data},
                        },
                    ]
                )

            check_scripts.extend(main_refs_checks)

        failure_items = []
        for check_script in check_scripts:
            try:
                # self.review_manager.logger.info(check_script["script"])
                if not check_script["params"]:
                    # self.review_manager.logger.debug(
                    #     "%s() called", check_script["script"].__name__
                    # )
                    check_script["script"]()
                else:
                    # self.review_manager.logger.debug(
                    #     "%s(params) called", check_script["script"].__name__
                    # )
                    if isinstance(check_script["params"], list):
                        check_script["script"](*check_script["params"])
                    else:
                        check_script["script"](**check_script["params"])
                # self.review_manager.logger.debug(
                #     "%s: passed\n", check_script["script"].__name__
                # )
            except (
                colrev_exceptions.MissingDependencyError,
                colrev_exceptions.GitConflictError,
                colrev_exceptions.PropagatedIDChange,
                colrev_exceptions.DuplicateIDsError,
                colrev_exceptions.OriginError,
                colrev_exceptions.FieldValueError,
                colrev_exceptions.StatusTransitionError,
                colrev_exceptions.UnstagedGitChangesError,
                colrev_exceptions.StatusFieldValueError,
            ) as exc:
                failure_items.append(f"{type(exc).__name__}: {exc}")
        return failure_items

    def check_repo(self) -> dict:
        """Check whether the repository is in a consistent state
        Entrypoint for pre-commit hooks
        """

        failure_items = []
        failure_items.extend(self.check_repo_extended())
        failure_items.extend(self.check_repo_basics())

        if failure_items:
            return {"status": ExitCodes.FAIL, "msg": "  " + "\n  ".join(failure_items)}
        return {"status": ExitCodes.SUCCESS, "msg": "Everything ok."}



#! /usr/bin/env python
"""Checker for name-format-titles."""
from __future__ import annotations

import re

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class NameFormatTitleChecker:
    """The NameFormatTitleChecker"""

    fields_to_check = ["author", "editor"]
    titles = ["MD", "Dr", "PhD", "Prof", "Dipl Ing"]
    __words_rgx = re.compile(r"(\w[\w']*\w|\w)")

    msg = "name-format-titles"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the name-format-titles checks"""

        for key in self.fields_to_check:
            if key not in record.data:
                continue

            if self.__title_in_name(name=record.data[key]):
                record.add_masterdata_provenance_note(key=key, note=self.msg)
            else:
                record.remove_masterdata_provenance_note(key=key, note=self.msg)

    def __title_in_name(self, *, name: str) -> bool:
        name_parts = self.__words_rgx.findall(name.lower().replace(".", ""))
        return any(title.lower() in name_parts for title in self.titles)


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(NameFormatTitleChecker(quality_model))



#! /usr/bin/env python
"""Resolution of BibTeX crossref fields as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods

if TYPE_CHECKING:
    import colrev.ops.prep
    import colrev.env.local_index


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class BibTexCrossrefResolutionPrep(JsonSchemaMixin):
    """Prepares records by resolving BibTex crossref links (e.g., to proceedings)"""

    settings_class = colrev.env.package_manager.DefaultSettings

    source_correction_hint = "check with the developer"
    always_apply_changes = False
    ci_supported: bool = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.review_manager = prep_operation.review_manager

    def __get_crossref_record(self, *, record: colrev.record.PrepRecord) -> dict:
        """Get the record linked through the BiBTex crossref field"""

        # Note : the ID of the crossrefed record_dict may have changed.
        # we need to trace based on the colrev_origin
        for crossref_origin in record.data["colrev_origin"]:
            crossref_origin = crossref_origin[: crossref_origin.rfind("/")]
            crossref_origin = crossref_origin + "/" + record.data["crossref"]
            for candidate_record_dict in self.review_manager.dataset.read_next_record():
                if crossref_origin in candidate_record_dict["colrev_origin"]:
                    return candidate_record_dict
        return {}

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare the record by resolving BiBTex crossref links (proceedings)"""

        if "crossref" not in record.data:
            return record

        crossref_record = self.__get_crossref_record(record=record)

        if not crossref_record:
            return record

        for key, value in crossref_record.items():
            if key not in record.data:
                record.update_field(
                    key=key,
                    value=value,
                    source="crossref_resolution",
                    append_edit=False,
                )
        del record.data["crossref"]

        return record



#! /usr/bin/env python
"""Convenience functions to load bib files"""
from __future__ import annotations

import os
import re
import typing
from typing import TYPE_CHECKING

import colrev.exceptions as colrev_exceptions

if TYPE_CHECKING:
    import colrev.ops.load


def __apply_file_fixes(
    *,
    load_operation: colrev.ops.load.Load,
    source: colrev.settings.SearchSource,
) -> None:
    # pylint: disable=duplicate-code
    # pylint: disable=too-many-statements

    def fix_key(
        file: typing.IO, line: bytes, replacement_line: bytes, seekpos: int
    ) -> int:
        load_operation.review_manager.logger.info(
            f"Fix invalid key: \n{line.decode('utf-8')}"
            f"{replacement_line.decode('utf-8')}"
        )
        line = file.readline()
        remaining = line + file.read()
        file.seek(seekpos)
        file.write(replacement_line)
        seekpos = file.tell()
        file.flush()
        os.fsync(file)
        file.write(remaining)
        file.truncate()  # if the replacement is shorter...
        file.seek(seekpos)
        return seekpos

    if not source.filename.is_file():
        load_operation.review_manager.logger.debug(
            f"Did not find bib file {source.filename} "
        )
        return

    with open(source.filename, encoding="utf8") as bibtex_file:
        contents = bibtex_file.read()
        bib_r = re.compile(r"@.*{.*,", re.M)
        if len(re.findall(bib_r, contents)) == 0:
            load_operation.review_manager.logger.error(
                f"Not a bib file? {source.filename.name}"
            )
            raise colrev_exceptions.UnsupportedImportFormatError(source.filename)

    # Errors to fix before pybtex loading:
    # - set_incremental_ids (otherwise, not all records will be loaded)
    # - fix_keys (keys containing white spaces)
    record_ids: typing.List[str] = []
    with open(source.filename, "r+b") as file:
        seekpos = file.tell()
        line = file.readline()
        while line:
            if b"@" in line[:3]:
                current_id = line[line.find(b"{") + 1 : line.rfind(b",")]
                current_id_str = current_id.decode("utf-8").lstrip().rstrip()

                if any(x in current_id_str for x in [";"]):
                    replacement_line = re.sub(
                        r";",
                        r"_",
                        line.decode("utf-8"),
                    ).encode("utf-8")
                    seekpos = fix_key(file, line, replacement_line, seekpos)

                if current_id_str in record_ids:
                    next_id = (
                        load_operation.review_manager.dataset.generate_next_unique_id(
                            temp_id=current_id_str, existing_ids=record_ids
                        )
                    )
                    load_operation.review_manager.logger.info(
                        f"Fix duplicate ID: {current_id_str} >> {next_id}"
                    )

                    replacement_line = (
                        line.decode("utf-8")
                        .replace(current_id.decode("utf-8"), next_id)
                        .encode("utf-8")
                    )

                    line = file.readline()
                    remaining = line + file.read()
                    file.seek(seekpos)
                    file.write(replacement_line)
                    seekpos = file.tell()
                    file.flush()
                    os.fsync(file)
                    file.write(remaining)
                    file.truncate()  # if the replacement is shorter...
                    file.seek(seekpos)

                    record_ids.append(next_id)

                else:
                    record_ids.append(current_id_str)

            # Fix keys
            if re.match(r"^\s*[a-zA-Z0-9]+\s+[a-zA-Z0-9]+\s*\=", line.decode("utf-8")):
                replacement_line = re.sub(
                    r"(^\s*)([a-zA-Z0-9]+)\s+([a-zA-Z0-9]+)(\s*\=)",
                    r"\1\2_\3\4",
                    line.decode("utf-8"),
                ).encode("utf-8")
                seekpos = fix_key(file, line, replacement_line, seekpos)

            seekpos = file.tell()
            line = file.readline()


def load_bib_file(
    load_operation: colrev.ops.load.Load,
    source: colrev.settings.SearchSource,
    check_bib_file: bool = True,
) -> dict:
    """Load a bib file and return records dict"""

    def drop_empty_fields(*, records: dict) -> None:
        for record_id in records:
            records[record_id] = {
                k: v for k, v in records[record_id].items() if v is not None
            }
            records[record_id] = {
                k: v for k, v in records[record_id].items() if v != "nan"
            }

    def check_nr_in_bib(*, source: colrev.settings.SearchSource, records: dict) -> None:
        load_operation.review_manager.logger.debug(
            f"Loaded {source.filename.name} with {len(records)} records"
        )
        nr_in_bib = load_operation.review_manager.dataset.get_nr_in_bib(
            file_path=source.filename
        )
        if len(records) < nr_in_bib:
            load_operation.review_manager.logger.error(
                "broken bib file (not imported all records)"
            )
            with open(source.filename, encoding="utf8") as file:
                line = file.readline()
                while line:
                    if "@" in line[:3]:
                        record_id = line[line.find("{") + 1 : line.rfind(",")]
                        if record_id not in [x["ID"] for x in records]:
                            load_operation.review_manager.logger.error(
                                f"{record_id} not imported"
                            )
                    line = file.readline()

    def __check_bib_file(
        *, source: colrev.settings.SearchSource, records: dict
    ) -> None:
        if len(records.items()) <= 3:
            return
        if not any("author" in r for r in records.values()):
            raise colrev_exceptions.ImportException(
                f"Import failed (no record with author field): {source.filename.name}"
            )

        if not any("title" in r for ID, r in records.items()):
            raise colrev_exceptions.ImportException(
                f"Import failed (no record with title field): {source.filename.name}"
            )

    def __load_records(*, source: colrev.settings.SearchSource) -> dict:
        if not source.filename.is_file():
            return {}
        with open(source.filename, encoding="utf8") as bibtex_file:
            records = load_operation.review_manager.dataset.load_records_dict(
                load_str=bibtex_file.read()
            )

            if len(records) == 0:
                load_operation.review_manager.report_logger.debug("No records loaded")
                load_operation.review_manager.logger.debug("No records loaded")
            return records

    def lower_case_keys(*, records: dict) -> None:
        for record in records.values():
            for key in list(record.keys()):
                if key in ["ID", "ENTRYTYPE"]:
                    continue
                if not key.islower():
                    record[key.lower()] = record.pop(key)

    __apply_file_fixes(load_operation=load_operation, source=source)

    records = __load_records(source=source)
    if len(records) == 0:
        return records

    lower_case_keys(records=records)
    drop_empty_fields(records=records)
    records = dict(sorted(records.items()))
    check_nr_in_bib(source=source, records=records)
    if check_bib_file:
        __check_bib_file(source=source, records=records)
    return records



#! /usr/bin/env python
"""Consolidation of metadata based on CiteAs API as a prep operation"""
from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class CiteAsPrep(JsonSchemaMixin):
    """Prepares records based on citeas.org metadata"""

    settings_class = colrev.env.package_manager.DefaultSettings

    source_correction_hint = "Search on https://citeas.org/ and click 'modify'"
    always_apply_changes = False
    ci_supported: bool = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.same_record_type_required = (
            prep_operation.review_manager.settings.is_curated_masterdata_repo()
        )
        self.session = prep_operation.review_manager.get_cached_session()
        _, self.email = prep_operation.review_manager.get_committer()

    def __cite_as_json_to_record(
        self, *, json_str: str, url: str
    ) -> colrev.record.PrepRecord:
        retrieved_record: dict = {}
        data = json.loads(json_str)

        if "author" in data["metadata"]:
            authors = data["metadata"]["author"]
            authors_string = ""
            for author in authors:
                authors_string += author.get("family", "") + ", "
                authors_string += author.get("given", "") + " "
            authors_string = authors_string.lstrip().rstrip().replace("  ", " ")
            retrieved_record.update(author=authors_string)
        if "container-title" in data["metadata"]:
            container_title = data["metadata"]["container-title"]
            if isinstance(data["metadata"]["container-title"], list):
                container_title = "".join(data["metadata"]["container-title"])
            retrieved_record.update(title=container_title)
        if "URL" in data["metadata"]:
            retrieved_record.update(url=data["metadata"]["URL"])
        if "note" in data["metadata"]:
            retrieved_record.update(note=data["metadata"]["note"])
        if "type" in data["metadata"]:
            retrieved_record.update(ENTRYTYPE=data["metadata"]["type"])
        if "year" in data["metadata"]:
            retrieved_record.update(year=data["metadata"]["year"])
        if "DOI" in data["metadata"]:
            retrieved_record.update(doi=data["metadata"]["DOI"])

        record = colrev.record.PrepRecord(data=retrieved_record)
        record.add_provenance_all(source=url)
        return record

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record based on citeas"""

        if record.data.get("ENTRYTYPE", "NA") not in ["misc", "software"]:
            return record
        if "title" not in record.data:
            return record

        try:
            url = (
                f"https://api.citeas.org/product/{record.data['title']}?"
                + f"email={self.email}"
            )
            ret = self.session.request(
                "GET",
                url,
                headers=prep_operation.requests_headers,
                timeout=prep_operation.timeout,
            )
            ret.raise_for_status()

            retrieved_record = self.__cite_as_json_to_record(json_str=ret.text, url=url)

            similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                record_original=retrieved_record,
                retrieved_record_original=retrieved_record,
                same_record_type_required=self.same_record_type_required,
            )
            if similarity > prep_operation.retrieval_similarity:
                record.merge(merging_record=retrieved_record, default_source=url)

        except (requests.exceptions.RequestException, colrev_exceptions.InvalidMerge):
            pass
        except UnicodeEncodeError:
            prep_operation.review_manager.logger.error(
                "UnicodeEncodeError - this needs to be fixed at some time"
            )

        return record



#! /usr/bin/env python
"""Template for a custom PDFPrep PackageEndpoint"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING

import zope.interface
from dacite import from_dict

import colrev.operation
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.pdf_prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PDFPrepPackageEndpointInterface)
class CustomPDFPrep:
    """Class for custom pdf-prep scripts"""

    def __init__(
        self,
        *,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = from_dict(
            data_class=colrev.env.package_manager.DefaultSettings, data=settings
        )

    def prep_pdf(
        self,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,  # pylint: disable=unused-argument
        record: colrev.record.Record,
        pad: int,  # pylint: disable=unused-argument
    ) -> colrev.record.Record:
        """Prepare the PDF"""

        if random.random() < 0.8:  # nosec
            record.add_data_provenance_note(key="file", note="custom_issue_detected")
            record.data.update(
                colrev_status=colrev.record.RecordState.pdf_needs_manual_preparation
            )

        return record



#!/usr/bin/env python3
"""Functionality for data/records.bib and git repository."""
from __future__ import annotations

import io
import itertools
import os
import re
import string
import time
import typing
from copy import deepcopy
from pathlib import Path
from random import randint
from typing import Optional
from typing import TYPE_CHECKING

import git
import pybtex.errors
from git.exc import GitCommandError
from git.exc import InvalidGitRepositoryError
from pybtex.database import Person
from pybtex.database.input import bibtex
from tqdm import tqdm

import colrev.env.language_service
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record
import colrev.settings

if TYPE_CHECKING:
    import colrev.review_manager

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines


class Dataset:
    """The CoLRev dataset (records and their history in git)"""

    RECORDS_FILE_RELATIVE = Path("data/records.bib")
    GIT_IGNORE_FILE_RELATIVE = Path(".gitignore")
    DEFAULT_GIT_IGNORE_ITEMS = [
        ".history",
        ".colrev",
        ".corrections",
        ".report.log",
        "__pycache__",
        "*.bib.sav",
        "venv",
        "output",
        "data/pdfs",
        "data/pdf_get_man/missing_pdf_files.csv",
        "data/.tei/",
        "data/prep_man/records_prep_man.bib",
        "data/prep/",
        "data/dedupe/",
    ]
    DEPRECATED_GIT_IGNORE_ITEMS = [
        "missing_pdf_files.csv",
        "manual_cleansing_statistics.csv",
        ".references_learned_settings",
        "pdfs",
        ".tei",
        "data.csv",
        "requests_cache.sqlite",
    ]

    records_file: Path
    __git_repo: git.Repo

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        self.review_manager = review_manager
        self.records_file = review_manager.path / self.RECORDS_FILE_RELATIVE
        self.git_ignore_file = review_manager.path / self.GIT_IGNORE_FILE_RELATIVE

        try:
            self.__git_repo = git.Repo(self.review_manager.path)
        except InvalidGitRepositoryError as exc:
            msg = "Not a CoLRev/git repository. Run\n    colrev init"
            raise colrev_exceptions.RepoSetupError(msg) from exc

        if not self.review_manager.verbose_mode:
            temp_f = io.StringIO()
            pybtex.io.stderr = temp_f

        self.update_gitignore(
            add=self.DEFAULT_GIT_IGNORE_ITEMS, remove=self.DEPRECATED_GIT_IGNORE_ITEMS
        )

    def update_gitignore(
        self, *, add: typing.Optional[list] = None, remove: typing.Optional[list] = None
    ) -> None:
        """Update the gitignore file by adding or removing particular paths"""
        # The following may print warnings...

        if self.git_ignore_file.is_file():
            gitignore_content = self.git_ignore_file.read_text(encoding="utf-8")
        else:
            gitignore_content = ""
        ignored_items = gitignore_content.splitlines()
        if remove:
            ignored_items = [x for x in ignored_items if x not in remove]
        if add:
            ignored_items = ignored_items + [
                str(a) for a in add if str(a) not in ignored_items
            ]

        with self.git_ignore_file.open("w", encoding="utf-8") as file:
            file.write("\n".join(ignored_items) + "\n")
        self.add_changes(path=self.git_ignore_file)

    def get_origin_state_dict(
        self, *, file_object: Optional[io.StringIO] = None
    ) -> dict:
        """Get the origin_state_dict (to determine state transitions efficiently)

        {'30_example_records.bib/Staehr2010': <RecordState.pdf_not_available: 10>,}
        """

        current_origin_states_dict = {}
        if self.records_file.is_file():
            for record_header_item in self.__read_record_header_items(
                file_object=file_object
            ):
                for origin in record_header_item["colrev_origin"]:
                    current_origin_states_dict[origin] = record_header_item[
                        "colrev_status"
                    ]
        return current_origin_states_dict

    def get_committed_origin_state_dict(self) -> dict:
        """Get the committed origin_state_dict"""

        filecontents = self.__get_last_records_filecontents()
        committed_origin_state_dict = self.get_origin_state_dict(
            file_object=io.StringIO(filecontents.decode("utf-8"))
        )
        return committed_origin_state_dict

    def get_nr_in_bib(self, *, file_path: Path) -> int:
        """Returns number of records in the bib file"""
        number_in_bib = 0
        with open(file_path, encoding="utf8") as file:
            line = file.readline()
            while line:
                if "@" in line[:3]:
                    if "@comment" not in line[:10].lower():
                        number_in_bib += 1
                line = file.readline()
        return number_in_bib

    def load_records_from_history(
        self, *, commit_sha: str = ""
    ) -> typing.Iterator[dict]:
        """Returns an iterator of the records_dict based on git history"""

        # If the records are not in the commit (commit_sha), we note that the
        # commit_sha was foud, but that the records were not changed in that commit.
        # It means that we ignore the StopIterations and
        # return the records from the next (prior) commit
        found_but_not_changed = False
        skipped_prior_commits = False  # if no commit_sha provided
        for commit in self.__git_repo.iter_commits():
            if commit_sha:
                if not skipped_prior_commits:
                    if not found_but_not_changed:
                        if commit_sha == commit.hexsha:
                            skipped_prior_commits = True
                        else:
                            continue

            try:
                filecontents = (
                    commit.tree / str(self.RECORDS_FILE_RELATIVE)
                ).data_stream.read()
                # Note : reinitialize parser (otherwise, bib_data does not change)
                parser = bibtex.Parser()
                bib_data = parser.parse_string(filecontents.decode("utf-8"))
                records_dict = self.parse_records_dict(records_dict=bib_data.entries)
            except (StopIteration, KeyError):
                found_but_not_changed = True
                continue
            yield records_dict

    def get_changed_records(self, *, target_commit: str) -> typing.List[dict]:
        """Get the records that changed in a selected commit"""

        revlist = (
            (
                commit.hexsha,
                (commit.tree / "data" / "records.bib").data_stream.read(),
            )
            for commit in self.__git_repo.iter_commits(
                paths=str(self.RECORDS_FILE_RELATIVE)
            )
        )
        found = False
        records_dict, prior_records_dict = {}, {}
        for commit, filecontents in list(revlist):
            if found:  # load the records_file_relative in the following commit
                prior_records_dict = self.review_manager.dataset.load_records_dict(
                    load_str=filecontents.decode("utf-8")
                )
                break
            if commit == target_commit:
                records_dict = self.review_manager.dataset.load_records_dict(
                    load_str=filecontents.decode("utf-8")
                )
                found = True

        # determine which records have been changed (prepared or merged)
        # in the target_commit
        for record in records_dict.values():
            prior_record_l = [
                rec
                for rec in prior_records_dict.values()
                if any(x in record["colrev_origin"] for x in rec["colrev_origin"])
            ]
            if not prior_record_l:
                continue
            prior_record = prior_record_l[0]
            # Note: the following is an exact comparison of all fields
            if record != prior_record:
                record.update(changed_in_target_commit="True")

        return list(records_dict.values())

    @classmethod
    def __load_field_dict(cls, *, value: str, field: str) -> dict:
        # pylint: disable=too-many-branches

        return_dict = {}
        if field == "colrev_masterdata_provenance":
            if value[:7] == "CURATED":
                if value.count(";") == 0:
                    value += ";;"  # Note : temporary fix (old format)
                if value.count(";") == 1:
                    value += ";"  # Note : temporary fix (old format)

                if ":" in value:
                    source = value[value.find(":") + 1 : value[:-1].rfind(";")]
                else:
                    source = ""
                return_dict["CURATED"] = {
                    "source": source,
                    "note": "",
                }

            elif value != "":
                # Pybtex automatically replaces \n in fields.
                # For consistency, we also do that for header_only mode:
                if "\n" in value:
                    value = value.replace("\n", " ")
                items = [x.lstrip() + ";" for x in (value + " ").split("; ") if x != ""]

                for item in items:
                    key_source = item[: item[:-1].rfind(";")]
                    if ":" in key_source:
                        note = item[item[:-1].rfind(";") + 1 : -1]
                        key, source = key_source.split(":", 1)
                        # key = key.rstrip().lstrip()
                        return_dict[key] = {
                            "source": source,
                            "note": note,
                        }
                    else:
                        print(f"problem with masterdata_provenance_item {item}")

        elif field == "colrev_data_provenance":
            if value != "":
                # Note : pybtex replaces \n upon load
                for item in (value + " ").split("; "):
                    if item == "":
                        continue
                    item += ";"  # removed by split
                    key_source = item[: item[:-1].rfind(";")]
                    note = item[item[:-1].rfind(";") + 1 : -1]
                    if ":" in key_source:
                        key, source = key_source.split(":", 1)
                        return_dict[key] = {
                            "source": source,
                            "note": note,
                        }
                    else:
                        print(f"problem with data_provenance_item {item}")

        else:
            print(f"error loading dict_field: {field}")

        return return_dict

    @classmethod
    def parse_records_dict(cls, *, records_dict: dict) -> dict:
        """Parse a records_dict from pybtex to colrev standard"""

        def format_name(person: Person) -> str:
            def join(name_list: list) -> str:
                return " ".join([name for name in name_list if name])

            first = person.get_part_as_text("first")
            middle = person.get_part_as_text("middle")
            prelast = person.get_part_as_text("prelast")
            last = person.get_part_as_text("last")
            lineage = person.get_part_as_text("lineage")
            name_string = ""
            if last:
                name_string += join([prelast, last])
            if lineage:
                name_string += f", {lineage}"
            if first or middle:
                name_string += ", "
                name_string += join([first, middle])
            return name_string

        # Need to concatenate fields and persons dicts
        # but pybtex is still the most efficient solution.
        records_dict = {
            k: {
                **{"ID": k},
                **{"ENTRYTYPE": v.type},
                **dict(
                    {
                        # Cast status to Enum
                        k: colrev.record.RecordState[v] if ("colrev_status" == k)
                        # DOIs are case insensitive -> use upper case.
                        else v.upper() if ("doi" == k)
                        # Note : the following two lines are a temporary fix
                        # to converg colrev_origins to list items
                        else [el.rstrip().lstrip() for el in v.split(";") if "" != el]
                        if k == "colrev_origin"
                        else [el.rstrip() for el in (v + " ").split("; ") if "" != el]
                        if k in colrev.record.Record.list_fields_keys
                        else Dataset.__load_field_dict(value=v, field=k)
                        if k in colrev.record.Record.dict_fields_keys
                        else v
                        for k, v in v.fields.items()
                    }
                ),
                **dict(
                    {
                        k: " and ".join(format_name(person) for person in persons)
                        for k, persons in v.persons.items()
                    }
                ),
            }
            for k, v in records_dict.items()
        }

        return records_dict

    def __parse_k_v(self, item_string: str) -> tuple:
        try:
            if " = " in item_string:
                key, value = item_string.split(" = ", 1)
            else:
                key = "ID"
                value = item_string.split("{")[1]

            key = key.lstrip().rstrip()
            value = value.lstrip().rstrip().lstrip("{").rstrip("},")
            if key == "colrev_origin":
                value_list = value.replace("\n", "").split(";")
                value_list = [x.lstrip(" ").rstrip(" ") for x in value_list if x]
                return key, value_list
            if key == "colrev_status":
                return key, colrev.record.RecordState[value]
            if key == "colrev_masterdata_provenance":
                return key, self.__load_field_dict(value=value, field=key)
            if key == "file":
                return key, Path(value)
        except IndexError as exc:
            raise colrev_exceptions.BrokenFilesError(msg="parsing records.bib") from exc

        return key, value

    def __read_record_header_items(
        self, *, file_object: Optional[typing.TextIO] = None
    ) -> list:
        # Note : more than 10x faster than the pybtex part of load_records_dict()

        # pylint: disable=consider-using-with
        if file_object is None:
            file_object = open(self.records_file, encoding="utf-8")

        # Fields required
        default = {
            "ID": "NA",
            "colrev_origin": "NA",
            "colrev_status": "NA",
            "file": "NA",
            "screening_criteria": "NA",
            "colrev_masterdata_provenance": "NA",
        }
        number_required_header_items = len(default)

        record_header_item = default.copy()
        item_count, item_string, record_header_items = (
            0,
            "",
            [],
        )
        while True:
            line = file_object.readline()
            if not line:
                break

            if line[:1] == "%" or line == "\n":
                continue

            if item_count > number_required_header_items or "}" == line:
                record_header_items.append(record_header_item)
                record_header_item = default.copy()
                item_count = 0
                continue

            if "@" in line[:2] and record_header_item["ID"] != "NA":
                record_header_items.append(record_header_item)
                record_header_item = default.copy()
                item_count = 0

            item_string += line
            if "}," in line or "@" in line[:2]:
                key, value = self.__parse_k_v(item_string)
                if key == "colrev_masterdata_provenance":
                    if value == "NA":
                        value = {}
                if value == "NA":
                    item_string = ""
                    continue
                item_string = ""
                if key in record_header_item:
                    item_count += 1
                    record_header_item[key] = value

        if record_header_item["colrev_origin"] != "NA":
            record_header_items.append(record_header_item)

        return [
            {k: v for k, v in record_header_item.items() if "NA" != v}
            for record_header_item in record_header_items
        ]

    def load_records_dict(
        self,
        *,
        file_path: Optional[Path] = None,
        load_str: Optional[str] = None,
        header_only: bool = False,
    ) -> dict:
        """Load the records

        - requires review_manager.notify(...)

        header_only:

        {"Staehr2010": {'ID': 'Staehr2010',
        'colrev_origin': ['30_example_records.bib/Staehr2010'],
        'colrev_status': <RecordState.md_imported: 2>,
        'screening_criteria': 'criterion1=in;criterion2=out',
        'file': PosixPath('data/pdfs/Smith2000.pdf'),
        'colrev_masterdata_provenance': {"author":{"source":"...", "note":"..."}}},
        }
        """

        if self.review_manager.notified_next_operation is None:
            raise colrev_exceptions.ReviewManagerNotNofiedError()

        pybtex.errors.set_strict_mode(False)
        if header_only:
            # Note : currently not parsing screening_criteria to settings.ScreeningCriterion
            # to optimize performance
            record_header_list = (
                self.__read_record_header_items() if self.records_file.is_file() else []
            )
            record_header_dict = {r["ID"]: r for r in record_header_list}
            return record_header_dict

        if file_path:
            with open(file_path, encoding="utf-8") as file:
                load_str = file.read()

        parser = bibtex.Parser()
        if load_str:
            # Fix missing comma after fields
            load_str = re.sub(r"(.)}\n", r"\g<1>},\n", load_str)
            bib_data = parser.parse_string(load_str)
            records_dict = self.parse_records_dict(records_dict=bib_data.entries)

        elif self.records_file.is_file():
            bib_data = parser.parse_file(str(self.records_file))
            records_dict = self.parse_records_dict(records_dict=bib_data.entries)
        else:
            records_dict = {}

        return records_dict

    @classmethod
    def parse_bibtex_str(
        cls,
        *,
        recs_dict_in: dict,
    ) -> str:
        """Parse a records_dict to a BiBTex string"""

        # Note: we need a deepcopy because the parsing modifies dicts
        recs_dict = deepcopy(recs_dict_in)

        def format_field(field: str, value: str) -> str:
            padd = " " * max(0, 28 - len(field))
            return f",\n   {field} {padd} = {{{value}}}"

        bibtex_str = ""
        language_service = colrev.env.language_service.LanguageService()
        first = True
        for record_id, record_dict in recs_dict.items():
            if not first:
                bibtex_str += "\n"
            first = False

            bibtex_str += f"@{record_dict['ENTRYTYPE']}{{{record_id}"

            try:
                language_service.unify_to_iso_639_3_language_codes(
                    record=colrev.record.Record(data=record_dict)
                )
            except colrev_exceptions.InvalidLanguageCodeException:
                del record_dict["language"]

            field_order = [
                "colrev_origin",  # must be in second line
                "colrev_status",
                "colrev_masterdata_provenance",
                "colrev_data_provenance",
                "colrev_pdf_id",
                "screening_criteria",
                "file",  # Note : do not change this order (parsers rely on it)
                "prescreen_exclusion",
                "doi",
                "grobid-version",
                "colrev.dblp.dblp_key",
                "colrev.semantic_scholar.id",
                "colrev.web_of_science.unique-id",
                "author",
                "booktitle",
                "journal",
                "title",
                "year",
                "volume",
                "number",
                "pages",
                "editor",
                "publisher",
                "url",
                "abstract",
            ]

            record = colrev.record.Record(data=record_dict)
            record_dict = record.get_data(stringify=True)

            for ordered_field in field_order:
                if ordered_field in record_dict:
                    if record_dict[ordered_field] == "":
                        continue
                    bibtex_str += format_field(
                        ordered_field, record_dict[ordered_field]
                    )

            for key in sorted(record_dict.keys()):
                if key in field_order + ["ID", "ENTRYTYPE"]:
                    continue

                bibtex_str += format_field(key, record_dict[key])

            bibtex_str += ",\n}\n"

        return bibtex_str

    def save_records_dict_to_file(self, *, records: dict, save_path: Path) -> None:
        """Save the records dict to specified file"""
        # Note : this classmethod function can be called by CoLRev scripts
        # operating outside a CoLRev repo (e.g., sync)

        bibtex_str = self.parse_bibtex_str(recs_dict_in=records)

        with open(save_path, "w", encoding="utf-8") as out:
            out.write(bibtex_str + "\n")

    def __save_record_list_by_id(
        self, *, records: dict, append_new: bool = False
    ) -> None:
        # Note : currently no use case for append_new=True??

        parsed = self.parse_bibtex_str(recs_dict_in=records)
        record_list = [
            {
                "ID": item[item.find("{") + 1 : item.find(",")],
                "record": "@" + item + "\n",
            }
            for item in parsed.split("\n@")
        ]
        # Correct the first item
        record_list[0]["record"] = "@" + record_list[0]["record"][2:]

        current_id_str = "NOTSET"
        if self.records_file.is_file():
            with open(self.records_file, "r+b") as file:
                seekpos = file.tell()
                line = file.readline()
                while line:
                    if b"@" in line[:3]:
                        current_id = line[line.find(b"{") + 1 : line.rfind(b",")]
                        current_id_str = current_id.decode("utf-8")
                    if current_id_str in [x["ID"] for x in record_list]:
                        replacement = [
                            x["record"]
                            for x in record_list
                            if x["ID"] == current_id_str
                        ][0]
                        record_list = [
                            x for x in record_list if x["ID"] != current_id_str
                        ]
                        line = file.readline()
                        while (
                            b"@" not in line[:3] and line
                        ):  # replace: drop the current record
                            line = file.readline()
                        remaining = line + file.read()
                        file.seek(seekpos)
                        file.write(replacement.encode("utf-8"))
                        seekpos = file.tell()
                        file.flush()
                        os.fsync(file)
                        file.write(remaining)
                        file.truncate()  # if the replacement is shorter...
                        file.seek(seekpos)

                    seekpos = file.tell()
                    line = file.readline()

        if len(record_list) > 0:
            if append_new:
                with open(self.records_file, "a", encoding="utf8") as m_refs:
                    for item in record_list:
                        m_refs.write(item["record"])
            else:
                self.review_manager.report_logger.error(
                    "records not written to file: " f'{[x["ID"] for x in record_list]}'
                )

        self.add_record_changes()

    def save_records_dict(self, *, records: dict, partial: bool = False) -> None:
        """Save the records dict in RECORDS_FILE"""

        if partial:
            self.__save_record_list_by_id(records=records)
            return
        self.save_records_dict_to_file(records=records, save_path=self.records_file)

    def read_next_record(
        self, *, conditions: Optional[list] = None
    ) -> typing.Iterator[dict]:
        """Read records (Iterator) based on condition"""

        # Note : matches conditions connected with 'OR'
        record_dict = self.load_records_dict()

        records = []
        for _, record in record_dict.items():
            if conditions is not None:
                for condition in conditions:
                    for key, value in condition.items():
                        if str(value) == str(record[key]):
                            records.append(record)
            else:
                records.append(record)
        yield from records

    def format_records_file(self) -> bool:
        """Format the records file"""
        quality_model = self.review_manager.get_qm()
        records = self.load_records_dict()
        for record_dict in records.values():
            if "colrev_status" not in record_dict:
                print(f'Error: no status field in record ({record_dict["ID"]})')
                continue

            record = colrev.record.PrepRecord(data=record_dict)
            if record_dict["colrev_status"] in [
                colrev.record.RecordState.md_needs_manual_preparation,
            ]:
                record.update_masterdata_provenance(qm=quality_model)
                record.update_metadata_status()

            if record_dict["colrev_status"] == colrev.record.RecordState.pdf_prepared:
                record.reset_pdf_provenance_notes()

        self.save_records_dict(records=records)
        changed = self.RECORDS_FILE_RELATIVE in [
            r.a_path for r in self.__git_repo.index.diff(None)
        ]
        return changed

    # ID creation, update and lookup ---------------------------------------

    def reprocess_id(self, *, paper_ids: str) -> None:
        """Remove an ID (set of IDs) from the bib_db (for reprocessing)"""

        saved_args = locals()
        if paper_ids == "all":
            # self.review_manager.logger.info("Removing/reprocessing all records")
            os.remove(self.records_file)
            self.__git_repo.index.remove(
                [str(self.RECORDS_FILE_RELATIVE)],
                working_tree=True,
            )
        else:
            records = self.load_records_dict()
            records = {
                ID: record
                for ID, record in records.items()
                if ID not in paper_ids.split(",")
            }
            self.save_records_dict(records=records)
            self.add_record_changes()

        self.review_manager.create_commit(msg="Reprocess", saved_args=saved_args)

    def __generate_temp_id(
        self, *, local_index: colrev.env.local_index.LocalIndex, record_dict: dict
    ) -> str:
        # pylint: disable=too-many-branches

        try:
            retrieved_record = local_index.retrieve(record_dict=record_dict)
            temp_id = retrieved_record["ID"]

            # Do not use IDs from local_index for curated_metadata repositories
            if "curated_metadata" in str(self.review_manager.path):
                raise colrev_exceptions.RecordNotInIndexException()

        except (
            colrev_exceptions.RecordNotInIndexException,
            colrev_exceptions.NotEnoughDataToIdentifyException,
        ):
            if record_dict.get("author", record_dict.get("editor", "")) != "":
                authors_string = record_dict.get(
                    "author", record_dict.get("editor", "Anonymous")
                )
                authors = colrev.record.PrepRecord.format_author_field(
                    input_string=authors_string
                ).split(" and ")
            else:
                authors = ["Anonymous"]

            # Use family names
            for author in authors:
                if "," in author:
                    author = author.split(",", maxsplit=1)[0]
                else:
                    author = author.split(" ", maxsplit=1)[0]

            id_pattern = self.review_manager.settings.project.id_pattern
            if colrev.settings.IDPattern.first_author_year == id_pattern:
                temp_id = (
                    f'{author.replace(" ", "")}{str(record_dict.get("year", "NoYear"))}'
                )
            elif colrev.settings.IDPattern.three_authors_year == id_pattern:
                temp_id = ""
                indices = len(authors)
                if len(authors) > 3:
                    indices = 3
                for ind in range(0, indices):
                    temp_id = temp_id + f'{authors[ind].split(",")[0].replace(" ", "")}'
                if len(authors) > 3:
                    temp_id = temp_id + "EtAl"
                temp_id = temp_id + str(record_dict.get("year", "NoYear"))

            if temp_id.isupper():
                temp_id = temp_id.capitalize()
            # Replace special characters
            # (because IDs may be used as file names)
            temp_id = colrev.env.utils.remove_accents(input_str=temp_id)
            temp_id = re.sub(r"\(.*\)", "", temp_id)
            temp_id = re.sub("[^0-9a-zA-Z]+", "", temp_id)

        return temp_id

    def generate_next_unique_id(
        self,
        *,
        temp_id: str,
        existing_ids: list,
    ) -> str:
        """Get the next unique ID"""

        order = 0
        letters = list(string.ascii_lowercase)
        next_unique_id = temp_id
        appends: list = []
        while next_unique_id.lower() in [i.lower() for i in existing_ids]:
            if len(appends) == 0:
                order += 1
                appends = list(itertools.product(letters, repeat=order))
            next_unique_id = temp_id + "".join(list(appends.pop(0)))
        temp_id = next_unique_id
        return temp_id

    def propagated_id(self, *, record_id: str) -> bool:
        """Check whether an ID is propagated (i.e., its record's status is beyond md_processed)"""

        for record in self.load_records_dict(header_only=True).values():
            if record["ID"] == record_id:
                if record[
                    "colrev_status"
                ] in colrev.record.RecordState.get_post_x_states(
                    state=colrev.record.RecordState.md_processed
                ):
                    return True

        return False

    def __generate_id(
        self,
        *,
        local_index: colrev.env.local_index.LocalIndex,
        record_dict: dict,
        existing_ids: Optional[list] = None,
    ) -> str:
        """Generate a blacklist to avoid setting duplicate IDs"""

        # Only change IDs that are before md_processed
        if record_dict["colrev_status"] in colrev.record.RecordState.get_post_x_states(
            state=colrev.record.RecordState.md_processed
        ):
            raise colrev_exceptions.PropagatedIDChange([record_dict["ID"]])
        # Alternatively, we could change IDs except for those
        # that have been propagated to the
        # screen or data will not be replaced
        # (this would break the chain of evidence)

        temp_id = self.__generate_temp_id(
            local_index=local_index, record_dict=record_dict
        )

        if existing_ids:
            temp_id = self.generate_next_unique_id(
                temp_id=temp_id,
                existing_ids=existing_ids,
            )

        return temp_id

    def set_ids(
        self, *, records: Optional[dict] = None, selected_ids: Optional[list] = None
    ) -> dict:
        """Set the IDs of records according to predefined formats or
        according to the LocalIndex"""
        # pylint: disable=redefined-outer-name

        local_index = self.review_manager.get_local_index()

        if records is None:
            records = {}

        if len(records) == 0:
            records = self.load_records_dict()

        id_list = list(records.keys())

        for record_id in tqdm(list(records.keys())):
            try:
                record_dict = records[record_id]
                if selected_ids is not None:
                    if record_id not in selected_ids:
                        continue
                if (
                    record_dict["colrev_status"]
                    not in [
                        colrev.record.RecordState.md_imported,
                        colrev.record.RecordState.md_prepared,
                    ]
                    and not self.review_manager.force_mode
                ):
                    continue
                old_id = record_id

                temp_stat = record_dict["colrev_status"]
                if selected_ids:
                    record = colrev.record.Record(data=record_dict)
                    record.set_status(
                        target_state=colrev.record.RecordState.md_prepared
                    )
                new_id = self.__generate_id(
                    local_index=local_index,
                    record_dict=record_dict,
                    existing_ids=[x for x in id_list if x != record_id],
                )
                if selected_ids:
                    record = colrev.record.Record(data=record_dict)
                    record.set_status(target_state=temp_stat)

                id_list.append(new_id)
                if old_id != new_id:
                    # We need to insert the a new element into records
                    # to make sure that the IDs are actually saved
                    record_dict.update(ID=new_id)
                    records[new_id] = record_dict
                    del records[old_id]
                    self.review_manager.report_logger.info(
                        f"set_ids({old_id}) to {new_id}"
                    )
                    if old_id in id_list:
                        id_list.remove(old_id)
            except colrev_exceptions.PropagatedIDChange as exc:
                print(exc)

        self.save_records_dict(records=records)
        self.add_record_changes()

        return records

    def get_next_id(self, *, bib_file: Path) -> int:
        """Get the next ID (incrementing counter)"""
        ids = []
        if bib_file.is_file():
            with open(bib_file, encoding="utf8") as file:
                line = file.readline()
                while line:
                    if "@" in line[:3]:
                        current_id = line[line.find("{") + 1 : line.rfind(",")]
                        ids.append(current_id)
                    line = file.readline()
        max_id = max([int(cid) for cid in ids if cid.isdigit()] + [0]) + 1
        return max_id

    # GIT operations -----------------------------------------------

    def get_repo(self) -> git.Repo:
        """Get the git repository object (requires review_manager.notify(...))"""

        if self.review_manager.notified_next_operation is None:
            raise colrev_exceptions.ReviewManagerNotNofiedError()
        return self.__git_repo

    def has_changes(
        self, *, relative_path: Optional[Path] = None, change_type: str = "all"
    ) -> bool:
        """Check whether the relative path (or the git repository) has changes"""

        if relative_path:
            main_recs_changed = False
            try:
                if change_type == "all":
                    main_recs_changed = str(relative_path) in [
                        item.a_path for item in self.__git_repo.index.diff(None)
                    ] + [item.a_path for item in self.__git_repo.head.commit.diff()]
                elif change_type == "staged":
                    main_recs_changed = str(relative_path) in [
                        item.a_path for item in self.__git_repo.head.commit.diff()
                    ]

                elif change_type == "unstaged":
                    main_recs_changed = str(relative_path) in [
                        item.a_path for item in self.__git_repo.index.diff(None)
                    ]
            except ValueError:
                pass
            return main_recs_changed

        return self.__git_repo.is_dirty()

    def add_changes(self, *, path: Path, remove: bool = False) -> None:
        """Add changed file to git"""

        if path.is_absolute():
            path = path.relative_to(self.review_manager.path)

        while (self.review_manager.path / Path(".git/index.lock")).is_file():
            time.sleep(randint(1, 50) * 0.1)  # nosec
            print("Waiting for previous git operation to complete")

        try:
            if remove:
                self.__git_repo.index.remove([str(path)])
            else:
                self.__git_repo.index.add([str(path)])
        except GitCommandError:
            pass

    def get_untracked_files(self) -> list:
        """Get the files that are untracked by git"""

        return self.__git_repo.untracked_files

    def __get_last_records_filecontents(self) -> bytes:
        revlist = (
            (
                commit.hexsha,
                (commit.tree / "data" / "records.bib").data_stream.read(),
            )
            for commit in self.__git_repo.iter_commits(
                paths=str(self.RECORDS_FILE_RELATIVE)
            )
        )
        filecontents = list(revlist)[0][1]
        return filecontents

    def records_changed(self) -> bool:
        """Check whether the records were changed"""
        try:
            main_recs_changed = str(self.RECORDS_FILE_RELATIVE) in [
                item.a_path for item in self.__git_repo.index.diff(None)
            ] + [x.a_path for x in self.__git_repo.head.commit.diff()]
            self.__get_last_records_filecontents()
        except (IndexError, ValueError, KeyError):
            main_recs_changed = False
        return main_recs_changed

    def remove_file_from_git(self, *, path: str) -> None:
        """Remove a file from git"""
        self.__git_repo.index.remove([path], working_tree=True)

    def create_commit(
        self, *, msg: str, author: git.Actor, committer: git.Actor, hook_skipping: bool
    ) -> None:
        """Create a commit"""
        self.__git_repo.index.commit(
            msg,
            author=author,
            committer=committer,
            skip_hooks=hook_skipping,
        )

    def file_in_history(self, *, filepath: Path) -> bool:
        """Check whether a file is in the git history"""
        return str(filepath) in [
            o.path for o in self.__git_repo.head.commit.tree.traverse()
        ]

    def get_commit_message(self, *, commit_nr: int) -> str:
        """Get the commit message for commit #"""
        master = self.__git_repo.head.reference
        assert commit_nr == 0  # extension : implement other cases
        if commit_nr == 0:
            cmsg = master.commit.message
        return cmsg

    def add_record_changes(self) -> None:
        """Add changes in records to git"""
        while (self.review_manager.path / Path(".git/index.lock")).is_file():
            time.sleep(randint(1, 50) * 0.1)  # nosec
            print("Waiting for previous git operation to complete")
        self.__git_repo.index.add([str(self.RECORDS_FILE_RELATIVE)])

    def add_setting_changes(self) -> None:
        """Add changes in settings to git"""
        while (self.review_manager.path / Path(".git/index.lock")).is_file():
            time.sleep(randint(1, 50) * 0.1)  # nosec
            print("Waiting for previous git operation to complete")

        self.__git_repo.index.add([str(self.review_manager.SETTINGS_RELATIVE)])

    def has_untracked_search_records(self) -> bool:
        """Check whether there are untracked search records"""
        search_dir = str(self.review_manager.SEARCHDIR_RELATIVE) + "/"
        untracked_files = self.get_untracked_files()
        return any(search_dir in untracked_file for untracked_file in untracked_files)

    def reset_log_if_no_changes(self) -> None:
        """Reset the report log file if there are not changes"""
        if not self.__git_repo.is_dirty():
            self.review_manager.reset_report_logger()

    def get_last_commit_sha(self) -> str:  # pragma: no cover
        """Get the last commit sha"""
        return str(self.__git_repo.head.commit.hexsha)

    def get_tree_hash(self) -> str:  # pragma: no cover
        """Get the current tree hash"""
        tree_hash = self.__git_repo.git.execute(["git", "write-tree"])
        return str(tree_hash)

    def __get_remote_commit_differences(self) -> list:  # pragma: no cover
        origin = self.__git_repo.remotes.origin
        if origin.exists():
            try:
                origin.fetch()
            except GitCommandError:
                return [-1, -1]

        nr_commits_behind, nr_commits_ahead = -1, -1
        if self.__git_repo.active_branch.tracking_branch() is not None:
            branch_name = str(self.__git_repo.active_branch)
            tracking_branch_name = str(self.__git_repo.active_branch.tracking_branch())
            # self.review_manager.logger.debug(f"{branch_name} - {tracking_branch_name}")

            behind_operation = branch_name + ".." + tracking_branch_name
            commits_behind = self.__git_repo.iter_commits(behind_operation)
            nr_commits_behind = sum(1 for c in commits_behind)

            ahead_operation = tracking_branch_name + ".." + branch_name
            commits_ahead = self.__git_repo.iter_commits(ahead_operation)
            nr_commits_ahead = sum(1 for c in commits_ahead)

        return [nr_commits_behind, nr_commits_ahead]

    def behind_remote(self) -> bool:  # pragma: no cover
        """Check whether the repository is behind the remote"""
        nr_commits_behind = 0
        connected_remote = 0 != len(self.__git_repo.remotes)
        if connected_remote:
            origin = self.__git_repo.remotes.origin
            if origin.exists():
                (
                    nr_commits_behind,
                    _,
                ) = self.__get_remote_commit_differences()
        if nr_commits_behind > 0:
            return True
        return False

    def remote_ahead(self) -> bool:  # pragma: no cover
        """Check whether the remote is ahead"""
        connected_remote = 0 != len(self.__git_repo.remotes)
        if connected_remote:
            origin = self.__git_repo.remotes.origin
            if origin.exists():
                (
                    _,
                    nr_commits_ahead,
                ) = self.__get_remote_commit_differences()
        if nr_commits_ahead > 0:
            return True
        return False

    def pull_if_repo_clean(self) -> None:  # pragma: no cover
        """Pull project if repository is clean"""
        if not self.__git_repo.is_dirty():
            origin = self.__git_repo.remotes.origin
            origin.pull()

    def get_remote_url(self) -> str:  # pragma: no cover
        """Get the remote url"""
        remote_url = "NA"
        for remote in self.__git_repo.remotes:
            if remote.name == "origin":
                remote_url = remote.url
        return remote_url

    def stash_unstaged_changes(self) -> bool:
        """Stash unstaged changes"""
        return "No local changes to save" != self.__git_repo.git.stash(
            "push", "--keep-index"
        )



#! /usr/bin/env python
"""Consolidation of metadata based on DOI metadata as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.built_in.search_sources.doi_org as doi_connector
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class DOIMetadataPrep(JsonSchemaMixin):
    """Prepares records based on doi.org metadata"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = (
        "ask the publisher to correct the metadata"
        + " (see https://www.crossref.org/blog/"
        + "metadata-corrections-updates-and-additions-in-metadata-manager/"
    )
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record by retrieving its metadata from doi.org"""

        if "doi" not in record.data:
            return record
        doi_connector.DOIConnector.retrieve_doi_metadata(
            review_manager=prep_operation.review_manager,
            record=record,
            timeout=prep_operation.timeout,
        )
        doi_connector.DOIConnector.get_link_from_doi(
            record=record,
            review_manager=prep_operation.review_manager,
        )
        return record



#! /usr/bin/env python
"""SearchSource: Springer Link"""
from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.load_utils_table
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code

# Note : API requires registration
# https://dev.springernature.com/


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class SpringerLinkSearchSource(JsonSchemaMixin):
    """SearchSource for Springer Link"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "Springer Link"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/springer_link.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.quality_model = source_operation.review_manager.get_qm()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Springer Link"""

        result = {"confidence": 0.0}

        if filename.suffix == ".csv":
            if data.count("http://link.springer.com") > data.count("\n") - 2:
                result["confidence"] = 1.0
                return result

        # Note : no features in bib file for identification

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of SpringerLink"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".csv":
            csv_loader = colrev.ops.load_utils_table.CSVLoader(
                load_operation=load_operation, source=self.search_source
            )
            records = csv_loader.load()
            self.__load_fixes(records=records)
            return records

        raise NotImplementedError

    def __load_fixes(
        self,
        records: typing.Dict,
    ) -> None:
        """Load fixes for Springer Link"""

        # pylint: disable=too-many-branches

        for record_dict in records.values():
            if "item_title" in record_dict:
                record_dict["title"] = record_dict["item_title"]
                del record_dict["item_title"]

            if record_dict.get("book_series_title", "") == "nan":
                del record_dict["book_series_title"]

            if "content_type" in record_dict:
                record = colrev.record.Record(data=record_dict)
                if record_dict["content_type"] == "Article":
                    if "publication_title" in record_dict:
                        record_dict["journal"] = record_dict["publication_title"]
                        del record_dict["publication_title"]
                    record.change_entrytype(
                        new_entrytype="article", qm=self.quality_model
                    )

                if record_dict["content_type"] == "Book":
                    if "publication_title" in record_dict:
                        record_dict["series"] = record_dict["publication_title"]
                        del record_dict["publication_title"]
                    record.change_entrytype(new_entrytype="book", qm=self.quality_model)

                if record_dict["content_type"] == "Chapter":
                    record_dict["chapter"] = record_dict["title"]
                    if "publication_title" in record_dict:
                        record_dict["title"] = record_dict["publication_title"]
                        del record_dict["publication_title"]
                    record.change_entrytype(
                        new_entrytype="inbook", qm=self.quality_model
                    )

                del record_dict["content_type"]

            if "item_doi" in record_dict:
                record_dict["doi"] = record_dict["item_doi"]
                del record_dict["item_doi"]
            if "journal_volume" in record_dict:
                record_dict["volume"] = record_dict["journal_volume"]
                del record_dict["journal_volume"]
            if "journal_issue" in record_dict:
                record_dict["number"] = record_dict["journal_issue"]
                del record_dict["journal_issue"]

            # Fix authors
            if "author" in record_dict:
                # a-bd-z: do not match McDonald
                record_dict["author"] = re.sub(
                    r"([a-bd-z]{1})([A-Z]{1})",
                    r"\g<1> and \g<2>",
                    record_dict["author"],
                )

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Springer Link"""

        return record



#!/usr/bin/env python3
"""Repare CoLRev projects."""
from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.operation

if TYPE_CHECKING:
    import colrev.review_manager


# pylint: disable=too-few-public-methods


class Repare(colrev.operation.Operation):
    """Repare a CoLRev project"""

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.check,
            notify_state_transition_operation=False,
        )
        # fix file no longer available
        self.local_index = self.review_manager.get_local_index()
        self.pdf_get_operation = self.review_manager.get_pdf_get_operation(
            notify_state_transition_operation=False
        )

    def __fix_broken_symlink_based_on_local_index(
        self, *, record: colrev.record.Record, full_path: Path
    ) -> None:
        """Fix broken symlinks based on local_index"""

        if os.path.islink(full_path) and not os.path.exists(full_path):
            self.review_manager.logger.debug(f" remove broken symlink: {full_path}")
            full_path.unlink()

        try:
            retrieved_record = self.local_index.retrieve(
                record_dict=record.data, include_file=True
            )

            if "file" in retrieved_record:
                record.update_field(
                    key="file",
                    value=str(retrieved_record["file"]),
                    source="local_index",
                    append_edit=False,
                )
                self.pdf_get_operation.import_pdf(record=record)
                if "fulltext" in retrieved_record:
                    del retrieved_record["fulltext"]
                self.review_manager.logger.info(
                    f" fix broken symlink: {record.data['ID']}"
                )
            else:
                self.review_manager.logger.debug(
                    f" file not linked in retrieved record: {record.data['ID']}"
                )

        except colrev_exceptions.RecordNotInIndexException:
            self.review_manager.logger.debug(
                f" record not in index: {record.data['ID']}"
            )

    def __fix_files(self, *, records: dict) -> None:
        for record_dict in records.values():
            if "file" not in record_dict:
                continue

            if not record_dict["file"].startswith("data/pdfs/"):
                record_dict["file"] = f"data/pdfs/{record_dict['ID']}.pdf"

            full_path = self.review_manager.path / Path(record_dict["file"])

            if full_path.is_file():
                continue

            # Add .pdf extension if missing
            if Path(str(full_path) + ".pdf").is_file():
                Path(str(full_path) + ".pdf").rename(full_path)

            # Check / replace multiple blanks in file and filename
            try:
                parent_dir = full_path.parent
                same_dir_pdfs = [
                    x.relative_to(self.review_manager.path)
                    for x in parent_dir.glob("*.pdf")
                ]
                for same_dir_pdf in same_dir_pdfs:
                    if record_dict["file"].replace("  ", " ") == str(
                        same_dir_pdf
                    ).replace("  ", " "):
                        same_dir_pdf.rename(str(same_dir_pdf).replace("  ", " "))
                        record_dict["file"] = record_dict["file"].replace("  ", " ")
            except ValueError:
                pass

            full_path = self.review_manager.path / Path(record_dict["file"])

            if not full_path.is_file():
                record = colrev.record.Record(data=record_dict)
                self.__fix_broken_symlink_based_on_local_index(
                    record=record, full_path=full_path
                )

            if full_path.is_file():
                continue

            record_dict["colrev_status_backup"] = record_dict["colrev_status"]
            del record_dict["file"]
            record = colrev.record.Record(data=record_dict)
            record.set_status(
                target_state=colrev.record.RecordState.rev_prescreen_included
            )

    def __get_source_feeds(self) -> dict:
        source_feeds = {}
        for source in self.review_manager.settings.sources:
            source_feeds[
                str(source.filename).replace("data/search/", "")
            ] = source.get_feed(
                review_manager=self.review_manager,
                source_identifier="NA",
                update_only=False,
            ).feed_records
        return source_feeds

    def __remove_fields(self, *, record: colrev.record.Record) -> None:
        if "pdf_hash" in record.data:
            record.data["colrev_pdf_id"] = record.data["pdf_hash"]
            del record.data["pdf_hash"]
        if "pdf_prep_hints" in record.data:
            del record.data["pdf_prep_hints"]
        if "grobid-version" in record.data:
            del record.data["grobid-version"]

        if "colrev_data_provenance" in record.data:
            mk_to_remove = []
            for key in record.data["colrev_data_provenance"]:
                if key not in record.data:
                    mk_to_remove += [key]
            for key in mk_to_remove:
                del record.data["colrev_data_provenance"][key]

        if "colrev_masterdata_provenance" in record.data:
            mdk_to_remove = []
            for key in record.data["colrev_masterdata_provenance"]:
                if (
                    key not in record.data
                    and "CURATED" != key
                    and "not-missing"
                    not in record.data["colrev_masterdata_provenance"][key]["note"]
                ):
                    mdk_to_remove += [key]
            for key in mdk_to_remove:
                del record.data["colrev_masterdata_provenance"][key]

        if self.review_manager.settings.is_curated_masterdata_repo():
            if "CURATED" in record.data["colrev_masterdata_provenance"]:
                del record.data["colrev_masterdata_provenance"]["CURATED"]

    def __set_data_provenance_field(
        self, *, record: colrev.record.Record, key: str, source_feeds: dict
    ) -> None:
        if key in record.data["colrev_data_provenance"]:
            if not any(
                record.data["colrev_data_provenance"][key]["source"].startswith(sf)
                for sf in list(source_feeds.keys())
            ):
                del record.data["colrev_data_provenance"][key]
        if key not in record.data["colrev_data_provenance"]:
            for origin in record.data["colrev_origin"]:
                origin_source = origin.split("/")[0]
                origin_id = origin[len(origin_source) + 1 :]
                if key in source_feeds[origin_source][origin_id]:
                    record.add_data_provenance(key=key, source=origin, note="")
        if key == "language":
            record.add_data_provenance(key=key, source="LanguageDetector", note="")
        if key == "tei_file":
            record.add_data_provenance(key=key, source="file|grobid", note="")
        if key == "colrev_pdf_id":
            record.add_data_provenance(key=key, source="file|pdf_hash", note="")

        if key not in record.data["colrev_data_provenance"]:
            record.add_data_provenance(key=key, source="manual", note="")

        for _, prov_details in record.data["colrev_data_provenance"].items():
            if prov_details["source"] in record.data["colrev_origin"] + ["manual"]:
                continue
            # Note : simple heuristic
            prov_details["source"] = record.data["colrev_origin"][0]

    def __add_missing_masterdata_provenance(
        self, *, record: colrev.record.Record, key: str, value: str, source_feeds: dict
    ) -> None:
        options = {}
        for origin in record.data["colrev_origin"]:
            origin_source = origin.split("/")[0]
            origin_id = origin[len(origin_source) + 1 :]
            if key in source_feeds[origin_source][origin_id]:
                options[origin_source] = source_feeds[origin_source][origin_id][key]

        source_value = "manual"

        # Note : simple heuristics:
        if value in options.values():
            if "md_curated.bib" in options:
                source_origin = "md_curated.bib"
            elif "md_crossref.bib" in options:
                source_origin = "md_crossref.bib"
            elif "CROSSREF.bib" in options:
                source_origin = "CROSSREF.bib"
            elif "md_dblp.bib" in options:
                source_origin = "md_dblp.bib"
            elif "DBLP.bib" in options:
                source_origin = "DBLP.bib"
            else:
                source_origin = [k for k, v in options.items() if v == value][0]

            for origin in record.data["colrev_origin"]:
                origin_source, origin_id = origin.split("/")
                if source_origin == origin_source:
                    source_value = origin

        record.add_masterdata_provenance(key=key, source=source_value, note="")

    def __set_non_curated_masterdata_provenance_field(
        self, *, record: colrev.record.Record, key: str, value: str, source_feeds: dict
    ) -> None:
        if key in record.data["colrev_masterdata_provenance"]:
            if not any(
                record.data["colrev_masterdata_provenance"][key]["source"].startswith(
                    sf
                )
                for sf in list(source_feeds.keys())
            ):
                del record.data["colrev_masterdata_provenance"][key]
        if key not in record.data["colrev_masterdata_provenance"]:
            self.__add_missing_masterdata_provenance(
                record=record, key=key, value=value, source_feeds=source_feeds
            )

        for prov_details in record.data["colrev_masterdata_provenance"].values():
            if prov_details["source"] in record.data["colrev_origin"] + ["manual"]:
                continue
            # Note : simple heuristic
            prov_details["source"] = record.data["colrev_origin"][0]

    def __set_provenance_field(
        self, *, record: colrev.record.Record, key: str, value: str, source_feeds: dict
    ) -> None:
        if key in colrev.record.Record.identifying_field_keys:
            if "CURATED" in record.data["colrev_masterdata_provenance"]:
                return
            self.__set_non_curated_masterdata_provenance_field(
                record=record, key=key, value=value, source_feeds=source_feeds
            )

        else:
            self.__set_data_provenance_field(
                record=record, key=key, source_feeds=source_feeds
            )

    def __set_provenance(
        self, *, record: colrev.record.Record, source_feeds: dict
    ) -> None:
        if "colrev_data_provenance" not in record.data:
            record.data["colrev_data_provenance"] = {}
        if "colrev_masterdata_provenance" not in record.data:
            record.data["colrev_masterdata_provenance"] = {}

        for key, value in record.data.items():
            if key in [
                "colrev_status",
                "colrev_origin",
                "colrev_masterdata_provenance",
                "colrev_data_provenance",
                "colrev_pdfid",
                "colrev_pdf_id",
                "colrev_id",
                "ID",
                "ENTRYTYPE",
                "screening_criteria",
            ]:
                continue
            self.__set_provenance_field(
                record=record, key=key, value=value, source_feeds=source_feeds
            )

    def __fix_provenance(self, *, records: dict) -> None:
        source_feeds = self.__get_source_feeds()
        for record_dict in records.values():
            record = colrev.record.Record(data=record_dict)
            self.__remove_fields(record=record)
            self.__set_provenance(record=record, source_feeds=source_feeds)

    def __fix_curated_sources(self, *, records: dict) -> None:
        local_index = self.review_manager.get_local_index()
        for search_source in self.review_manager.settings.sources:
            if search_source.endpoint != "colrev.local_index":
                continue
            curation_recs = self.review_manager.dataset.load_records_dict(
                file_path=search_source.filename
            )
            for record_id in list(curation_recs.keys()):
                if "curation_ID" not in curation_recs[record_id]:
                    try:
                        retrieved_record_dict = local_index.retrieve(
                            record_dict=curation_recs[record_id], include_file=False
                        )
                        del retrieved_record_dict["colrev_status"]
                        curation_recs[record_id] = retrieved_record_dict
                    except colrev_exceptions.RecordNotInIndexException:
                        main_record_origin = (
                            search_source.get_origin_prefix() + "/" + record_id
                        )
                        main_record_l = [
                            r
                            for r in records.values()
                            if main_record_origin in r["colrev_origin"]
                        ]
                        if not main_record_l:
                            continue
                        main_record_id = main_record_l[0]["ID"]
                        if 1 == len(records[main_record_id]["colrev_origin"]):
                            del records[main_record_id]
                        else:
                            records[main_record_id]["colrev_origin"].remove(
                                main_record_origin
                            )
                        del curation_recs[record_id]

            self.review_manager.dataset.save_records_dict_to_file(
                records=curation_recs, save_path=search_source.filename
            )
            self.review_manager.dataset.add_changes(path=search_source.filename)

    def __update_field_names(self, *, records: dict) -> None:
        for record_dict in records.values():
            # TODO : extract to methods and call from upgrade.py
            # TBD: which parts are in upgrade/repare and which parts are in prepare??
            record = colrev.record.Record(data=record_dict)
            if "fulltext" in record_dict.get("link", ""):
                record.rename_field(key="link", new_key="fulltext")
            if (
                record.data.get("note", "").lower().startswith("cited by ")
                and "cited_by" not in record.data
            ):
                record.data["note"] = record.data["note"][9:]
                record.rename_field(key="note", new_key="cited_by")
            if (
                record.data.get("note", "").lower().startswith("cited by ")
                and "cited_by" in record.data
            ):
                record.remove_field(key="note")
            if record.data.get("link", "").startswith(
                "https://api.elsevier.com/content/article/"
            ) and record.data.get("link", "").endswith("Accept=text/xml"):
                record.remove_field(key="link")

    @colrev.operation.Operation.decorate()
    def main(self) -> None:
        """Repare a CoLRev project (main entrypoint)"""

        # Try: open settings, except: notify & start repare

        # ...

        # Try: open records, except: notify & start repare
        try:
            records = self.review_manager.dataset.load_records_dict()
        except AttributeError:
            self.review_manager.logger.error("Could not read bibtex file")

            separated_records = {}  # type: ignore
            with open(
                self.review_manager.dataset.records_file, encoding="utf-8"
            ) as file:
                record_str = ""
                line = file.readline()

                while line:
                    if line == "\n":
                        records = self.review_manager.dataset.load_records_dict(
                            load_str=record_str
                        )
                        if len(records) != 1:
                            print(record_str)
                        else:
                            separated_records = {**separated_records, **records}
                        record_str = ""
                    record_str += line
                    line = file.readline()
            self.review_manager.dataset.save_records_dict_to_file(
                records=separated_records, save_path=Path("extracted.bib")
            )
            try:
                records = self.review_manager.dataset.load_records_dict()
            except AttributeError:
                return

        self.__fix_curated_sources(records=records)

        # removing specific fields
        # for record_dict in records.values():
        #     if "colrev_status_backup" in record_dict:
        #         colrev.record.Record(data=record_dict).remove_field(
        #             key="colrev_status_backup"
        #         )
        #     if "colrev_local_index" in record_dict:
        #         colrev.record.Record(data=record_dict).remove_field(
        #             key="colrev_local_index"
        #         )

        self.__update_field_names(records=records)

        self.__fix_provenance(records=records)

        self.__fix_files(records=records)

        self.review_manager.dataset.save_records_dict(records=records)
